#include <ioavr.h>              
#include <inavr.h>
#include "main.h"
#include "TWI_Master.h"

static uchar TWI_buf[ TWI_BUFFER_SIZE ];                                        // Transceiver buffer
static uchar TWI_msgSize;                                                       // Number of bytes to be transmitted.
static uchar TWI_state = TWI_NO_STATE;                                          // State byte. Default set to TWI_NO_STATE.

union TWI_statusReg TWI_statusReg = {0};                                        // TWI_statusReg is defined in TWI_Master.h

/*******************************************************************************
*                                                                              *
*******************************************************************************/
void TWI_Master_Initialise(void)
{
  TWBR = TWI_REG_SPEED;                                                         // Set bit rate register (Baudrate). Defined in header file.
  TWSR = TWI_BIT_DIV;                                                           // Driver presumes prescaler.
  TWDR = 0xFF;                                                                  // Default content = SDA released.
  TWCR = (1 << TWEN)                                                            // Enable TWI-interface and release TWI pins.
       | (0 << TWIE) | (0 << TWINT)                                             // Disable Interupt.
       | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                              // No Signal requests.
       | (0 << TWWC);
}    
/*******************************************************************************
*                                                                              *
*******************************************************************************/
uchar TWI_Transceiver_Busy(void)
{
  return ( TWCR & (1<<TWIE) );                                                  // IF TWI Interrupt is enabled then the Transceiver is busy
}
/*******************************************************************************
*                                                                              *
*******************************************************************************/
uchar TWI_Get_State_Info(void)
{
    while ( TWI_Transceiver_Busy() );                                           // Wait until TWI has completed the transmission.
    return ( TWI_state );                                                       // Return error state.
}
/*******************************************************************************
*                                                                              *
*******************************************************************************/
void TWI_Start_Transceiver_With_Data(uchar *msg, uchar msgSize)
{
  uchar temp;

    while ( TWI_Transceiver_Busy() );                                             // Wait until TWI is ready for next transmission.

    TWI_msgSize = msgSize;                                                        // Number of data to transmit.
    TWI_buf[0]  = msg[0];                                                         // Store slave address with R/W setting.

    if (!( msg[0] & (TRUE<<TWI_READ_BIT) ))                                       // If it is a write operation, then also copy data.
    {
        for ( temp = 1; temp < msgSize; temp++ )
        {
          TWI_buf[ temp ] = msg[ temp ];
        }
    }

    TWI_statusReg.all = 0;      
    TWI_state         = TWI_NO_STATE ;
    TWCR = (1 << TWEN)                                                          // TWI Interface enabled.
         | (1 << TWIE) | (1 << TWINT)                                           // Enable TWI Interupt and clear the flag.
         | (0 << TWEA) | (1 << TWSTA) | (0 << TWSTO)                            // Initiate a START condition.
         | (0 << TWWC);
}
/*******************************************************************************
*                                                                              *
*******************************************************************************/
void TWI_Start_Transceiver(void)
{
    while ( TWI_Transceiver_Busy() );                                           // Wait until TWI is ready for next transmission.
    TWI_statusReg.all = 0;      
    TWI_state         = TWI_NO_STATE;
    TWCR = (1 << TWEN)                                                          // TWI Interface enabled.
         | (1 << TWIE) | (1 << TWINT)                                           // Enable TWI Interupt and clear the flag.
         | (0 << TWEA) | (1 << TWSTA) | (0 << TWSTO)                            // Initiate a START condition.
         | (0 << TWWC);
}
/*******************************************************************************
*                                                                              *
*******************************************************************************/
uchar TWI_Get_Data_From_Transceiver(uchar *msg, uchar msgSize)
{
  uchar i;

    while ( TWI_Transceiver_Busy() );                                           // Wait until TWI is ready for next transmission.

    if( TWI_statusReg.lastTransOK )                                             // Last transmission competed successfully.              
    {                                             
        for ( i = 1; i < msgSize; i++ )                                         // Copy data from Transceiver buffer.
        {
          msg[ i-1 ] = TWI_buf[ i ];                                            //     
        }
    }

    return( TWI_statusReg.lastTransOK );                                   
}
/*******************************************************************************
*******************************************************************************/
#pragma vector=TWI_vect
__interrupt void TWI_ISR(void)
{
  static uchar   TWI_bufPtr;
  
    switch (TWSR & 0xF8)
    {
      case TWI_START:                                                           // START has been transmitted  
      case TWI_REP_START:                                                       // Repeated START has been transmitted
        TWI_bufPtr = 0;                                                         // Set buffer pointer to the TWI Address location
      case TWI_MTX_ADR_ACK:                                                     // SLA+W has been tramsmitted and ACK received
      case TWI_MTX_DATA_ACK:                                                    // Data byte has been tramsmitted and ACK received
        if (TWI_bufPtr < TWI_msgSize)
        {
          TWDR = TWI_buf[TWI_bufPtr++];
          TWCR = (1 << TWEN)                                                    // TWI Interface enabled
               | (1 << TWIE) | (1 << TWINT)                                     // Enable TWI Interupt and clear the flag to send byte
               | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO)
               | (0 << TWWC);
        }

        else                                                                    // Send STOP after last byte
        {
          TWI_statusReg.lastTransOK = TRUE;                                     // Set status bits to completed successfully. 
          TWCR = (1 << TWEN)                                                    // TWI Interface enabled
               | (0 << TWIE) | (1 << TWINT)                                     // Disable TWI Interrupt and clear the flag
               | (0 << TWEA) | (0 << TWSTA) | (1 << TWSTO)                      // Initiate a STOP condition.
               | (0 << TWWC);
        }
      break;

      case TWI_MRX_DATA_ACK:                                                    // Data byte has been received and ACK tramsmitted
        TWI_buf[TWI_bufPtr++] = TWDR;
      case TWI_MRX_ADR_ACK:                                                     // SLA+R has been tramsmitted and ACK received
        if (TWI_bufPtr < (TWI_msgSize-1) )                                      // Detect the last byte to NACK it.
        {
          TWCR = (1 << TWEN)                                                    // TWI Interface enabled
               | (1 << TWIE) | (1 << TWINT)                                     // Enable TWI Interupt and clear the flag to read next byte
               | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                      // Send ACK after reception
               | (0 << TWWC);
        }
      
        else                                                                    // Send NACK after next reception
        {
          TWCR = (1 << TWEN)                                                    // TWI Interface enabled
               | (1 << TWIE) | (1 << TWINT)                                     // Enable TWI Interupt and clear the flag to read next byte
               | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                      // Send NACK after reception
               | (0 << TWWC);
        }    
      break; 

      case TWI_MRX_DATA_NACK:                                                   // Data byte has been received and NACK tramsmitted
        TWI_buf[TWI_bufPtr] = TWDR;
        TWI_statusReg.lastTransOK = TRUE;                                       // Set status bits to completed successfully. 
        TWCR = (1 << TWEN)                                                      // TWI Interface enabled
             | (0 << TWIE) | (1 << TWINT)                                       // Disable TWI Interrupt and clear the flag
             | (0 << TWEA) | (0 << TWSTA) | (1 << TWSTO)                        // Initiate a STOP condition.
             | (0 << TWWC);
      break;      

      case TWI_ARB_LOST:                                                        // Arbitration lost
        TWCR = (1 << TWEN)                                                      // TWI Interface enabled
             | (1 << TWIE) | (1 << TWINT)                                       // Enable TWI Interupt and clear the flag
             | (0 << TWEA) | (1 << TWSTA) | (0 << TWSTO)                        // Initiate a (RE)START condition.
             | (0 << TWWC);
      break;

      case TWI_MTX_ADR_NACK:                                                    // SLA+W has been tramsmitted and NACK received
      case TWI_MRX_ADR_NACK:                                                    // SLA+R has been tramsmitted and NACK received    
      case TWI_MTX_DATA_NACK:                                                   // Data byte has been tramsmitted and NACK received
//      case TWI_NO_STATE                                                       // No relevant state information available; TWINT = 0
      case TWI_BUS_ERROR:                                                       // Bus error due to an illegal START or STOP condition
      default:     
        TWI_state = TWSR & 0xF8;                                                // Store TWSR and automatically sets clears noerrors bit.
                                                                                // Reset TWI Interface
        TWCR = (1 << TWEN)                                                      // Enable TWI-interface and release TWI pins
             | (0 << TWIE) | (0 << TWINT)                                       // Disable Interupt
             | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                        // No Signal requests
             | (0 << TWWC);
    }
}
