#include <ioavr.h>              
#include <inavr.h>
#include "main.h"
#include "TWI_slave.h"
 
static uchar TWI_buf[TWI_BUFFER_SIZE];                                          // Transceiver buffer. Set the size in the header file
static uchar TWI_msgSize  = 0;                                                  // Number of bytes to be transmitted.
static uchar TWI_state    = TWI_NO_STATE;                                       // State byte. Default set to TWI_NO_STATE.

union TWI_statusReg TWI_statusReg = {0};                                        // TWI_statusReg is defined in TWI_Slave.h

/*******************************************************************************
*******************************************************************************/
void TWI_Slave_Initialise( uchar TWI_ownAddress )
{
    TWAR = TWI_ownAddress;                                                      // Set own TWI slave address. Accept TWI General Calls.
    TWDR = 0xFF;                                                                // Default content = SDA released.
    TWCR = (1 << TWEN)                                                          // Enable TWI-interface and release TWI pins.
         | (0 << TWIE) | (0 << TWINT)                                           // Disable TWI Interupt.
         | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                            // Do not ACK on any requests, yet.
         | (0 << TWWC);
}    
/*******************************************************************************
*******************************************************************************/
uchar TWI_Transceiver_Busy( void )
{
    return ( TWCR & (1 << TWIE) );                                              // IF TWI interrupt is enabled then the Transceiver is busy
}
/*******************************************************************************
*******************************************************************************/
uchar TWI_Get_State_Info( void )
{
    while ( TWI_Transceiver_Busy() );                                           // Wait until TWI has completed the transmission.
    return ( TWI_state );                                                       // Return error state. 
}
/*******************************************************************************
*******************************************************************************/
void TWI_Start_Transceiver_With_Data( uchar *msg, uchar msgSize )
{
  uchar temp;

    while ( TWI_Transceiver_Busy() );                                           // Wait until TWI is ready for next transmission.

    TWI_msgSize = msgSize;                                                      // Number of data to transmit.

    for ( temp = 0; temp < msgSize; temp++ )                                    // Copy data that may be transmitted if the TWI Master requests data.
    {
      TWI_buf[ temp ] = msg[ temp ];
    }

    TWI_statusReg.all = 0;      
    TWI_state         = TWI_NO_STATE ;
    TWCR = (1 << TWEN)                                                          // TWI Interface enabled.
         | (1 << TWIE) | (1 << TWINT)                                           // Enable TWI Interupt and clear the flag.
         | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                            // Prepare to ACK next time the Slave is addressed.
         | (0 << TWWC);
}
/*******************************************************************************
*******************************************************************************/
void TWI_Start_Transceiver( void )
{
    while ( TWI_Transceiver_Busy() );                                           // Wait until TWI is ready for next transmission.
    TWI_statusReg.all = 0;      
    TWI_state         = TWI_NO_STATE ;
    TWCR = (1 << TWEN)                                                          // TWI Interface enabled.
         | (1 << TWIE) | (1 << TWINT)                                           // Enable TWI Interupt and clear the flag.
         | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                            // Prepare to ACK next time the Slave is addressed.
         | (0 << TWWC);
}
/*******************************************************************************
*******************************************************************************/
uchar TWI_Get_Data_From_Transceiver( uchar *msg, uchar msgSize )
{
  uchar i;

    while ( TWI_Transceiver_Busy() );                                           // Wait until TWI is ready for next transmission.

    if( TWI_statusReg.lastTransOK )                                             // Last transmission competed successfully.              
    {                                             
        for ( i = 0; i < msgSize; i++ )                                         // Copy data from Transceiver buffer.
        {
          msg[ i ] = TWI_buf[ i ];
        }

      TWI_statusReg.RxDataInBuf = FALSE;                                        // Slave Receive data has been read from buffer.
    }
    
    return( TWI_statusReg.lastTransOK );                                   
}
/*******************************************************************************
*******************************************************************************/
#pragma vector=TWI_vect
__interrupt void TWI_ISR( void )
{
   static uchar   TWI_bufPtr;
  
    switch (TWSR & 0xF8)
    {
      case TWI_STX_ADR_ACK:                                                     // Own SLA+R has been received; ACK has been returned
//      case TWI_STX_ADR_ACK_M_ARB_LOST:                                        // Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
        TWI_bufPtr   = 0;                                                       // Set buffer pointer to first data location
      case TWI_STX_DATA_ACK:                                                    // Data byte in TWDR has been transmitted; ACK has been received
        TWDR = TWI_buf[TWI_bufPtr++];
        TWCR = (1 << TWEN)                                                      // TWI Interface enabled
             | (1 << TWIE) | (1 << TWINT)                                       // Enable TWI Interupt and clear the flag to send byte
             | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO)
             | (0 << TWWC);
      break;

      case TWI_STX_DATA_NACK:                                                   // Data byte in TWDR has been transmitted; NACK has been received. 
                                                                                // I.e. this could be the end of the transmission.
        if (TWI_bufPtr == TWI_msgSize)                                          // Have we transceived all expected data?
        {
          TWI_statusReg.lastTransOK = TRUE;                                     // Set status bits to completed successfully. 
        }

        else                                                                    // Master has sent a NACK before all data where sent.
        {
          TWI_state = TWSR & 0xF8;                                              // Store TWI State as errormessage.      
        }        
                                                                                // Put TWI Transceiver in passive mode.
        TWCR = (1 << TWEN)                                                      // Enable TWI-interface and release TWI pins
             | (0 << TWIE) | (0 << TWINT)                                       // Disable Interupt
             | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                        // Do not acknowledge on any new requests.
             | (0 << TWWC);
      break;     

      case TWI_SRX_GEN_ACK:                                                     // General call address has been received; ACK has been returned
//      case TWI_SRX_GEN_ACK_M_ARB_LOST:                                        // Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
        TWI_statusReg.genAddressCall = TRUE;
      case TWI_SRX_ADR_ACK:                                                     // Own SLA+W has been received ACK has been returned
//      case TWI_SRX_ADR_ACK_M_ARB_LOST:                                        // Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned    
                                                                                // Dont need to clear TWI_S_statusRegister.generalAddressCall due to that it is the default state.
        TWI_statusReg.RxDataInBuf = TRUE;      
        TWI_bufPtr   = 0;                                                       // Set buffer pointer to first data location
                                                                                // Reset the TWI Interupt to wait for a new event.
        TWCR = (1 << TWEN)                                                      // TWI Interface enabled
             | (1 << TWIE) | (1 << TWINT)                                       // Enable TWI Interupt and clear the flag to send byte
             | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                        // Expect ACK on this transmission
             | (0 << TWWC);
      break;

      case TWI_SRX_ADR_DATA_ACK:                                                // Previously addressed with own SLA+W; data has been received; ACK has been returned
      case TWI_SRX_GEN_DATA_ACK:                                                // Previously addressed with general call; data has been received; ACK has been returned
        TWI_buf[TWI_bufPtr++]     = TWDR;
        TWI_statusReg.lastTransOK = TRUE;                                       // Set flag transmission successfull.       
                                                                                // Reset the TWI Interupt to wait for a new event.
        TWCR = (1 << TWEN)                                                      // TWI Interface enabled
             | (1 << TWIE) | (1 << TWINT)                                       // Enable TWI Interupt and clear the flag to send byte
             | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                        // Send ACK after next reception
             | (0 << TWWC);
      break;

      case TWI_SRX_STOP_RESTART:                                                // A STOP condition or repeated START condition has been received while still addressed as Slave    
                                                                                // Put TWI Transceiver in passive mode.
        TWCR = (1 << TWEN)                                                      // Enable TWI-interface and release TWI pins
             | (0 << TWIE) | (0 << TWINT)                                       // Disable Interupt
             | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                        // Do not acknowledge on any new requests.
             | (0 << TWWC);
      break;           

      case TWI_SRX_ADR_DATA_NACK:                                               // Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
      case TWI_SRX_GEN_DATA_NACK:                                               // Previously addressed with general call; data has been received; NOT ACK has been returned
      case TWI_STX_DATA_ACK_LAST_BYTE:                                          // Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received
//      case TWI_NO_STATE                                                       // No relevant state information available; TWINT = 0
      case TWI_BUS_ERROR:                                                       // Bus error due to an illegal START or STOP condition
      default:     
        TWI_state = TWSR & 0xF8;                                                // Store TWI State as errormessage, operation also clears the Success bit.      
        TWCR = (1 << TWEN)                                                      // Enable TWI-interface and release TWI pins
             | (0 << TWIE) | (0 << TWINT)                                       // Disable Interupt
             | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO)                        // Do not acknowledge on any new requests.
             | (0 << TWWC);
    }
}
